@@ -2,6 +2,34 @@
 
 extern struct uwsgi_server uwsgi;
 
+void *uwsgi_zeromq_init() {
+	if (!uwsgi.zmq_context) {
+		uwsgi.zmq_context = zmq_init(1);
+                if (uwsgi.zmq_context == NULL) {
+                        uwsgi_error("zmq_init()");
+                        exit(1);
+                }
+	}
+	return uwsgi.zmq_context;
+}
+
+void uwsgi_zeromq_init_sockets() {
+
+	uwsgi_zeromq_init();	
+
+        struct uwsgi_socket *uwsgi_sock = uwsgi.sockets;
+        while(uwsgi_sock) {
+                        if (!uwsgi_sock->proto_name || strcmp(uwsgi_sock->proto_name, "zmq")) {
+                                goto zmq_next;
+                        }
+                        uwsgi_proto_zeromq_setup(uwsgi_sock);
+zmq_next:
+                        uwsgi_sock = uwsgi_sock->next;
+
+                }
+
+}
+
 #ifdef UWSGI_JSON
 #include <jansson.h>
 
@@ -171,8 +199,8 @@ static int uwsgi_mongrel2_tnetstring_parse(struct wsgi_request *wsgi_req, char *
 		if (key[0] < 97) {
 			if (!uwsgi_strncmp("METHOD", 6, key, keylen)) {
 				if (!uwsgi_strncmp("JSON", 4, val, vallen)) {
-                                        return -1;
-                                }
+					return -1;
+				}
 				wsgi_req->uh.pktsize += proto_base_add_uwsgi_var(wsgi_req, "REQUEST_METHOD", 14, val, vallen);
 			}
 			else if (!uwsgi_strncmp("VERSION", 7, key, keylen)) {
@@ -253,6 +281,133 @@ int uwsgi_proto_zeromq_parser(struct wsgi_request *wsgi_req) {
 	return UWSGI_OK;
 }
 
+void uwsgi_proto_zeromq_thread_fixup(struct uwsgi_socket *uwsgi_sock, int async_id) {
+
+	void *tmp_zmq_pull = zmq_socket(uwsgi.zmq_context, ZMQ_PULL);
+        if (tmp_zmq_pull == NULL) {
+        	uwsgi_error("zmq_socket()");
+                exit(1);
+        }
+
+        if (zmq_connect(tmp_zmq_pull, uwsgi_sock->receiver) < 0) {
+        	uwsgi_error("zmq_connect()");
+                exit(1);
+        }
+
+	pthread_setspecific(uwsgi_sock->key, tmp_zmq_pull);
+
+#ifdef ZMQ_FD
+	if (uwsgi.threads > 1) {
+        	size_t zmq_socket_len = sizeof(int);
+        	if (zmq_getsockopt(pthread_getspecific(uwsgi_sock->key), ZMQ_FD, &uwsgi_sock->fd_threads[async_id], &zmq_socket_len) < 0) {
+        		uwsgi_error("zmq_getsockopt()");
+                	exit(1);
+		}
+		uwsgi_sock->retry[async_id] = 1;
+	}
+#endif
+}
+
+void uwsgi_proto_zeromq_setup(struct uwsgi_socket *uwsgi_sock) {
+
+	           char *responder = strchr(uwsgi_sock->name, ',');
+                        if (!responder) {
+                                uwsgi_log("invalid zeromq address\n");
+                                exit(1);
+                        }
+                        uwsgi_sock->receiver = uwsgi_concat2n(uwsgi_sock->name, responder - uwsgi_sock->name, "", 0);
+                        responder++;
+
+                        uwsgi_sock->pub = zmq_socket(uwsgi.zmq_context, ZMQ_PUB);
+                        if (uwsgi_sock->pub == NULL) {
+                                uwsgi_error("zmq_socket()");
+                                exit(1);
+                        }
+
+
+                        // generate uuid
+                        uuid_t uuid_zmq;
+                        uuid_generate(uuid_zmq);
+                        uuid_unparse(uuid_zmq, uwsgi_sock->uuid);
+
+                        if (zmq_setsockopt(uwsgi_sock->pub, ZMQ_IDENTITY, uwsgi_sock->uuid, 36) < 0) {
+                                uwsgi_error("zmq_setsockopt()");
+                                exit(1);
+                        }
+
+                        if (zmq_connect(uwsgi_sock->pub, responder) < 0) {
+                                uwsgi_error("zmq_connect()");
+                                exit(1);
+                        }
+
+                        uwsgi_log("zeromq UUID for responder %s on worker %d: %.*s\n", responder, uwsgi.mywid, 36, uwsgi_sock->uuid);
+
+                        uwsgi_sock->proto = uwsgi_proto_zeromq_parser;
+                        uwsgi_sock->proto_accept = uwsgi_proto_zeromq_accept;
+                        uwsgi_sock->proto_close = uwsgi_proto_zeromq_close;
+                        uwsgi_sock->proto_write = uwsgi_proto_zeromq_write;
+                        uwsgi_sock->proto_writev = uwsgi_proto_zeromq_writev;
+                        uwsgi_sock->proto_write_header = uwsgi_proto_zeromq_write_header;
+                        uwsgi_sock->proto_writev_header = uwsgi_proto_zeromq_writev_header;
+                        uwsgi_sock->proto_sendfile = uwsgi_proto_zeromq_sendfile;
+
+                        uwsgi_sock->proto_thread_fixup = uwsgi_proto_zeromq_thread_fixup;
+
+                        uwsgi_sock->edge_trigger = 1;
+                        uwsgi_sock->retry = uwsgi_malloc(sizeof(int) * uwsgi.threads);
+			uwsgi_sock->retry[0] = 1;
+
+                        // inform loop engine about edge trigger status
+                        uwsgi.is_et = 1;
+
+
+                        // initialize a lock for multithread usage
+                        if (uwsgi.threads > 1) {
+                                pthread_mutex_init(&uwsgi_sock->lock, NULL);
+                        }
+
+                        // one pull per-thread
+                        if (pthread_key_create(&uwsgi_sock->key, NULL)) {
+                                uwsgi_error("pthread_key_create()");
+                                exit(1);
+                        }
+
+                        void *tmp_zmq_pull = zmq_socket(uwsgi.zmq_context, ZMQ_PULL);
+                        if (tmp_zmq_pull == NULL) {
+                                uwsgi_error("zmq_socket()");
+                                exit(1);
+                        }
+                        if (zmq_connect(tmp_zmq_pull, uwsgi_sock->receiver) < 0) {
+                                uwsgi_error("zmq_connect()");
+                                exit(1);
+                        }
+
+                        pthread_setspecific(uwsgi_sock->key, tmp_zmq_pull);
+
+#ifdef ZMQ_FD
+                        size_t zmq_socket_len = sizeof(int);
+                        if (zmq_getsockopt(pthread_getspecific(uwsgi_sock->key), ZMQ_FD, &uwsgi_sock->fd, &zmq_socket_len) < 0) {
+                                uwsgi_error("zmq_getsockopt()");
+                                exit(1);
+                        }
+			if (uwsgi.threads > 1) {
+				uwsgi_sock->fd_threads = uwsgi_malloc(sizeof(int) * uwsgi.threads);
+				uwsgi_sock->fd_threads[0] = uwsgi_sock->fd;
+			}
+#else
+                        uwsgi_sock->fd = -1;
+#endif
+
+                        uwsgi_sock->bound = 1;
+#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(3,0,0)
+			uwsgi_sock->recv_flag = ZMQ_DONTWAIT;
+#else
+                        uwsgi_sock->recv_flag = ZMQ_NOBLOCK;
+#endif
+}
+
+
+
 int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 
 	zmq_msg_t message;
@@ -277,37 +432,36 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 
 #ifdef ZMQ_EVENTS
 	size_t events_len = sizeof(uint32_t);
-	if (uwsgi.edge_triggered == 0) {
-		if (zmq_getsockopt(pthread_getspecific(uwsgi.zmq_pull), ZMQ_EVENTS, &events, &events_len) < 0) {
-			uwsgi_error("zmq_getsockopt()");
-			uwsgi.edge_triggered = 0;
-			return -1;
-		}
+	if (zmq_getsockopt(pthread_getspecific(wsgi_req->socket->key), ZMQ_EVENTS, &events, &events_len) < 0) {
+		uwsgi_error("zmq_getsockopt()");
+		goto retry;
 	}
 #endif
 
-	if (events & ZMQ_POLLIN || uwsgi.edge_triggered) {
+	if (events & ZMQ_POLLIN || (wsgi_req->socket->retry && wsgi_req->socket->retry[wsgi_req->async_id])) {
 		wsgi_req->do_not_add_to_async_queue = 1;
 		wsgi_req->proto_parser_status = 0;
 		zmq_msg_init(&message);
-		if (zmq_recv(pthread_getspecific(uwsgi.zmq_pull), &message, uwsgi.zeromq_recv_flag) < 0) {
+#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(3,0,0)
+		if (zmq_recvmsg(pthread_getspecific(wsgi_req->socket->key), &message, wsgi_req->socket->recv_flag) < 0) {
+#else
+		if (zmq_recv(pthread_getspecific(wsgi_req->socket->key), &message, wsgi_req->socket->recv_flag) < 0) {
+#endif
 			if (errno == EAGAIN) {
-				uwsgi.edge_triggered = 0;
-			}
-			else {
-				uwsgi_error("zmq_recv()");
+				zmq_msg_close(&message);
+				goto repoll;
 			}
+			uwsgi_error("zmq_recv()");
 			zmq_msg_close(&message);
-			return -1;
+			goto retry;
 		}
-		uwsgi.edge_triggered = 1;
+
 		message_size = zmq_msg_size(&message);
 		//uwsgi_log("%.*s\n", (int) wsgi_req->proto_parser_pos, zmq_msg_data(&message));
 		if (message_size > 0xffff) {
 			uwsgi_log("too much big message %d\n", message_size);
 			zmq_msg_close(&message);
-			wsgi_req->do_not_log = 1;
-			return -1;
+			goto retry;
 		}
 
 		message_ptr = zmq_msg_data(&message);
@@ -317,8 +471,7 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 		if (post_data == NULL) {
 			uwsgi_log("cannot parse message (split4 phase)\n");
 			zmq_msg_close(&message);
-			wsgi_req->do_not_log = 1;
-			return -1;
+			goto retry;
 		}
 
 		// fix post_data, mongrel2_req and mongrel2_req_size
@@ -326,16 +479,14 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 		if (post_data == NULL) {
 			uwsgi_log("cannot parse message (body netstring phase)\n");
 			zmq_msg_close(&message);
-			wsgi_req->do_not_log = 1;
-			return -1;
+			goto retry;
 		}
 
 		// ok ready to parse tnetstring/json data and build uwsgi request
 		if (mongrel2_req[mongrel2_req_size] == '}') {
 			if (uwsgi_mongrel2_tnetstring_parse(wsgi_req, mongrel2_req, mongrel2_req_size)) {
 				zmq_msg_close(&message);
-				wsgi_req->do_not_log = 1;
-				return -1;
+				goto retry;
 			}
 		}
 		else {
@@ -349,15 +500,13 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 			if (!root) {
 				uwsgi_log("error parsing JSON data: line %d %s\n", error.line, error.text);
 				zmq_msg_close(&message);
-				wsgi_req->do_not_log = 1;
-				return -1;
+				goto retry;
 			}
 
 			if (uwsgi_mongrel2_json_parse(root, wsgi_req)) {
 				json_decref(root);
 				zmq_msg_close(&message);
-				wsgi_req->do_not_log = 1;
-				return -1;
+				goto retry;
 			}
 
 			json_decref(root);
@@ -388,7 +537,7 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 				if (fwrite(message_ptr, wsgi_req->post_cl, 1, wsgi_req->async_post) != 1) {
 					uwsgi_error("fwrite()");
 					zmq_msg_close(&message);
-					return -1;
+					goto retry;
 				}
 				rewind(wsgi_req->async_post);
 				wsgi_req->body_as_file = 1;
@@ -398,9 +547,20 @@ int uwsgi_proto_zeromq_accept(struct wsgi_request *wsgi_req, int fd) {
 
 		zmq_msg_close(&message);
 
+		// retry by default
+		wsgi_req->socket->retry[wsgi_req->async_id] = 1;
+
 		return 0;
 	}
 
+repoll:
+	// force polling of the socket
+	wsgi_req->socket->retry[wsgi_req->async_id] = 0;
+	return -1;
+retry:
+	// retry til EAGAIN;
+	wsgi_req->do_not_log = 1;
+	wsgi_req->socket->retry[wsgi_req->async_id] = 1;
 	return -1;
 }
 
@@ -416,11 +576,15 @@ void uwsgi_proto_zeromq_close(struct wsgi_request *wsgi_req) {
 		return;
 
 	zmq_msg_init_data(&reply, wsgi_req->proto_parser_buf, wsgi_req->proto_parser_pos, uwsgi_proto_zeromq_free, NULL);
-	if (uwsgi.threads > 1) pthread_mutex_lock(&uwsgi.zmq_lock);
-	if (zmq_send(uwsgi.zmq_pub, &reply, 0)) {
+	if (uwsgi.threads > 1) pthread_mutex_lock(&wsgi_req->socket->lock);
+#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(3,0,0)
+	if (zmq_sendmsg(wsgi_req->socket->pub, &reply, 0)) {
+#else
+	if (zmq_send(wsgi_req->socket->pub, &reply, 0)) {
+#endif
 		uwsgi_error("zmq_send()");
 	}
-	if (uwsgi.threads > 1) pthread_mutex_unlock(&uwsgi.zmq_lock);
+	if (uwsgi.threads > 1) pthread_mutex_unlock(&wsgi_req->socket->lock);
 	zmq_msg_close(&reply);
 
 	if (wsgi_req->async_post && wsgi_req->body_as_file) {
@@ -433,18 +597,24 @@ void uwsgi_proto_zeromq_close(struct wsgi_request *wsgi_req) {
 ssize_t uwsgi_proto_zeromq_writev_header(struct wsgi_request *wsgi_req, struct iovec *iovec, size_t iov_len) {
 	int i;
 	ssize_t len;
-	ssize_t ret = 0;
+
+	struct uwsgi_buffer *ub = uwsgi_buffer_new(4096);
 
 	for (i = 0; i < (int) iov_len; i++) {
-		len = uwsgi_proto_zeromq_write(wsgi_req, iovec[i].iov_base, iovec[i].iov_len);
-		if (len <= 0) {
+		if (uwsgi_buffer_append(ub, iovec[i].iov_base, iovec[i].iov_len)) {
 			wsgi_req->write_errors++;
 			return 0;
 		}
-		ret += len;
 	}
 
-	return ret;
+	len = uwsgi_proto_zeromq_write(wsgi_req, ub->buf, ub->pos);
+	if (len <= 0) {
+		wsgi_req->write_errors++;
+		return 0;
+	}
+
+	uwsgi_buffer_destroy(ub);
+	return len;
 }
 
 ssize_t uwsgi_proto_zeromq_writev(struct wsgi_request * wsgi_req, struct iovec * iovec, size_t iov_len) {
@@ -463,17 +633,21 @@ ssize_t uwsgi_proto_zeromq_write(struct wsgi_request * wsgi_req, char *buf, size
 	//uwsgi_log("|%.*s|\n", (int)wsgi_req->proto_parser_pos+len, zmq_body);
 
 	zmq_msg_init_data(&reply, zmq_body, wsgi_req->proto_parser_pos + len, uwsgi_proto_zeromq_free, NULL);
-	if (uwsgi.threads > 1) pthread_mutex_lock(&uwsgi.zmq_lock);
-	if (zmq_send(uwsgi.zmq_pub, &reply, 0)) {
+	if (uwsgi.threads > 1) pthread_mutex_lock(&wsgi_req->socket->lock);
+#if ZMQ_VERSION >= ZMQ_MAKE_VERSION(3,0,0)
+	if (zmq_sendmsg(wsgi_req->socket->pub, &reply, 0)) {
+#else
+	if (zmq_send(wsgi_req->socket->pub, &reply, 0)) {
+#endif
 		if (!uwsgi.ignore_write_errors) {
 			uwsgi_error("zmq_send()");
 		}
 		wsgi_req->write_errors++;
-		if (uwsgi.threads > 1) pthread_mutex_unlock(&uwsgi.zmq_lock);
+		if (uwsgi.threads > 1) pthread_mutex_unlock(&wsgi_req->socket->lock);
 		zmq_msg_close(&reply);
 		return 0;
 	}
-	if (uwsgi.threads > 1) pthread_mutex_unlock(&uwsgi.zmq_lock);
+	if (uwsgi.threads > 1) pthread_mutex_unlock(&wsgi_req->socket->lock);
 	zmq_msg_close(&reply);
 
 	return len;
